/*
Rafael Guglielmetti
http://www.allpotes.ch/raf
raf_guglielmetti@yahoo.fr
2004
*/

#include "cavalier.h"

// Si 1 : Affiche la grille aprs chaque tape
#define __DEBUG__ 0


/* 
		main()
		Fonction principale 
		Paramtres: Aucuns
		Retour: 0
*/
int main()
{
	// bienvenue
	cout << "--------------------------------------\n";
	cout << "Programme pour le probleme du cavalier\n";
	cout << "--------------------------------------\n";

	// affiche le menu
	menu();

	cout << "\nAppuyez sur une touche pour terminer le programme... ";
	getch();
	return 0;
}

/*
		menu()
		Affiche le menu, appelle les fonctions d'initialisation, de calcul et de terminaison de prog
		Paramtres: Aucuns
		Retour: Aucun
*/
void menu()
{
	// largeur, hauteur, dpart X, dpart Y
	int nWidth, nHeight, nStartX, nStartY;

	cout << "\nEntrez la largeur de votre echiquier: ";
	cin >> nWidth;

	cout << "\nEntrez la hauteur de votre echiquer: ";
	cin >> nHeight;

	// errreur de taille
	if(nWidth < 0 || nHeight < 0 || nWidth > 10 || nHeight > 10)
	{
		cout << "La taille que vous avez entre n'est pas valide...\n";
		return;
	}

	cout << "\nEntrez la coordonnee X de depart (de 0 a " << nWidth - 1 << ") : ";
	cin >> nStartX;

	cout << "\nEntrez la coordonnee X de depart (de 0 a " << nHeight - 1 << ") : ";
	cin >> nStartY;

	// erreur de coordonnes
	if(nStartX < 0 || nStartY < 0 || nStartX > nWidth - 1 || nStartY > nHeight - 1)
	{
		cout << "Les coordonnees de depart que vous avez entrees sont invalides\n";
		return;
	}

	// on tient compte de la bordure
	nStartX += 2;
	nStartY += 2;

	cout << "\n---------------------\nDebut du programmme...\nInitialisations: ";

	if(!initialisations(nWidth, nHeight, nStartX, nStartY))
	{
		cout << "Erreur\n";
		return;
	}

	cout << "OK\nDebut du calcul..." << endl;

	// lancement du calcul
	calculer(nWidth, nHeight);

	cout << "Liberation de la memoire: ";
	// affichage du rsultat, libratiuon de la mmoire, etc
	end_prog(nWidth, nHeight);
	cout << "Termine\n";
}

/*
		initialisations(int nWidth, int nHeight, int nStartX, int nStartY)
		Allocation de la mmoire, remplissage de l'echiquier
		Paramtres:
			- nWidth: Largeur de l'echiquier
			- nHeight: Hauteur de l'echiquier
			- nStartX: Position X de dpart
			- nStartY: Position Y de dpart
		Retour: true si succs, false sinon
*/
bool initialisations(int nWidth, int nHeight, int nStartX, int nStartY)
{
	// variables de boucles
	int i, j;

	// cration de l'echiquier
	nEchiquier = new int *[nHeight + 4];
	if(nEchiquier == NULL)
		return false;

	for(i = 0; i < nHeight + 4; i++)
		nEchiquier[i] = new int [nWidth + 4];

	// cration du tableau des positions
	nPos = new int *[nWidth * nHeight];
	if(nPos == NULL)
		return false;

	for(i = 0; i < nWidth * nHeight; i++)
		nPos[i] = new int [2];

	// remplissage du tableau (0 pour les cases de l'echiquier, -1 pour les cases de bordure)
	for(i = 0; i < nHeight + 4; i++)
	{
		// si 1e et 2e ligne || avant dernire et dernire
		if(i == 0 || i == 1 || i == nHeight + 2 || i == nHeight + 3)
			for(j = 0; j < nWidth + 4; j++)
				nEchiquier[i][j] = -1;
		else
		{
			nEchiquier[i][0] = nEchiquier[i][1] = nEchiquier[i][nWidth + 2] = nEchiquier[i][nWidth + 3] = -1;
			for(j = 2; j < nWidth + 2; j++)
				nEchiquier[i][j] = 0;
		}
	}

	// dpart
	nPos[0][0] = nStartX;
	nPos[0][1] = nStartY;

	return true;
}

/*
		void calculer(int nWidth, int nHeight)
		Allocation de la mmoire, remplissage de l'echiquier
		Paramtres:
			- nWidth: Largeur de l'echiquier
			- nHeight: Hauteur de l'echiquier
		Retour: Aucun
*/
void calculer(int nWidth, int nHeight)
{
	// coup courant et max, x courant, y courant, etat courant
	int nCurrent = 0, nMax = nWidth * nHeight - 1, nCurrentX, nCurrentY, nCurrentEtat;

	/*
			Code des dplacement

			| |8| |1| |
			|7| | | |2|
			| | |x| | |
			|6| | | |3|
			| |5| |4| |
	*/

	// boucle de calcul
	while(nCurrent < nMax)
	{
		// coordonnes courantes
		nCurrentX = nPos[nCurrent][0];
		nCurrentY = nPos[nCurrent][1];
		// etat courant
		nCurrentEtat = nEchiquier[nCurrentY][nCurrentX];

		// pas de trajet possible?
		if(nCurrent == 0)
		{
			if(nCurrentEtat == 0 && nEchiquier[nCurrentY - 2][nCurrentX + 1] == 0)
				;
			else if(nCurrentEtat < 2 && nEchiquier[nCurrentY - 1][nCurrentX + 2] == 0)
				;
			else if(nCurrentEtat < 3 && nEchiquier[nCurrentY + 1][nCurrentX + 2] == 0)
				;
			else if(nCurrentEtat < 4 && nEchiquier[nCurrentY + 2][nCurrentX + 1] == 0)
				;
			else if(nCurrentEtat < 5 && nEchiquier[nCurrentY + 2][nCurrentX - 1] == 0)
				;
			else if(nCurrentEtat < 6 && nEchiquier[nCurrentY + 1][nCurrentX - 2] == 0)
				;
			else if(nCurrentEtat < 7 && nEchiquier[nCurrentY - 1][nCurrentX - 2] == 0)
				;
			else if(nCurrentEtat < 8 && nEchiquier[nCurrentY - 2][nCurrentX - 1] == 0)
				;
			else
			{
				cout << "Cette configuration ne permet pas de solution...\n";
				return;
			}
		}

		// coup courant
		nCurrent++;

		// affichage de l'echiquier
	#if __DEBUG__ == 1
		write_echiquier(nWidth, nHeight, nCurrentX, nCurrentY);
	#endif

		// Dplacement 1
		if(nCurrentEtat == 0 && nEchiquier[nCurrentY - 2][nCurrentX + 1] == 0)
		{
			nEchiquier[nCurrentY][nCurrentX] = 1;
			nPos[nCurrent][0] = nCurrentX + 1;
			nPos[nCurrent][1] = nCurrentY - 2;
		}
		// Dplacement 2
		else if(nCurrentEtat < 2 && nEchiquier[nCurrentY - 1][nCurrentX + 2] == 0)
		{
			nEchiquier[nCurrentY][nCurrentX] = 2;
			nPos[nCurrent][0] = nCurrentX + 2;
			nPos[nCurrent][1] = nCurrentY - 1;
		}
		// Dplacement 3
		else if(nCurrentEtat < 3 && nEchiquier[nCurrentY + 1][nCurrentX + 2] == 0)
		{
			nEchiquier[nCurrentY][nCurrentX] = 3;
			nPos[nCurrent][0] = nCurrentX + 2;
			nPos[nCurrent][1] = nCurrentY + 1;
		}
		// Dplacement 4
		else if(nCurrentEtat < 4 && nEchiquier[nCurrentY + 2][nCurrentX + 1] == 0)
		{
			nEchiquier[nCurrentY][nCurrentX] = 4;
			nPos[nCurrent][0] = nCurrentX + 1;
			nPos[nCurrent][1] = nCurrentY + 2;
		}
		// Dplacement 5
		else if(nCurrentEtat < 5 && nEchiquier[nCurrentY + 2][nCurrentX - 1] == 0)
		{
			nEchiquier[nCurrentY][nCurrentX] = 5;
			nPos[nCurrent][0] = nCurrentX - 1;
			nPos[nCurrent][1] = nCurrentY + 2;
		}
		// Dplacement 6
		else if(nCurrentEtat < 6 && nEchiquier[nCurrentY + 1][nCurrentX - 2] == 0)
		{
			nEchiquier[nCurrentY][nCurrentX] = 6;
			nPos[nCurrent][0] = nCurrentX - 2;
			nPos[nCurrent][1] = nCurrentY + 1;
		}
		// Dplacement 7
		else if(nCurrentEtat < 7 && nEchiquier[nCurrentY - 1][nCurrentX - 2] == 0)
		{
			nEchiquier[nCurrentY][nCurrentX] = 7;
			nPos[nCurrent][0] = nCurrentX - 2;
			nPos[nCurrent][1] = nCurrentY - 1;
		}
		// Dplacement 8
		else if(nCurrentEtat < 8 && nEchiquier[nCurrentY - 2][nCurrentX - 1] == 0)
		{
			nEchiquier[nCurrentY][nCurrentX] = 8;
			nPos[nCurrent][0] = nCurrentX - 1;
			nPos[nCurrent][1] = nCurrentY - 2;
		}
		// On doit reculer
		else
		{
			nCurrent--;
			nEchiquier[nPos[nCurrent][1]][nPos[nCurrent][0]] = 0;

			do
			{
				nEchiquier[nPos[nCurrent][1]][nPos[nCurrent][0]] = 0;

				if(nCurrent == 0)
					break;

				nCurrent--;
			}
			while(nEchiquier[nPos[nCurrent][1]][nPos[nCurrent][0]] > 7);
		}
	}
}

/*
		void end_prog(int nWidth, int nHeight)
		Libration de la mmoire, affichage du rsultat + sortie fichier
		Paramtres:
			- nWidth: Largeur de l'echiquier
			- nHeight: Hauteur de l'echiquier
		Retour: Aucun
*/
void end_prog(int nWidth, int nHeight)
{
	int i, j;

	// affichage du resultat et sortie fichier
	cout << "\n\nParcours:\nX | Y\n";
	ofstream out;
	out.open("sortie.txt");
	for(i = 0; i < nWidth * nHeight; i++)
	{
		cout << nPos[i][0] - 2 << " | " << nPos[i][1] - 2 << endl;
		out << nPos[i][0] - 2 << " | " << nPos[i][1] - 2 << endl;
	}


	// affichage du trajet
	for(i = 0; i < nWidth * nHeight; i++)
		nEchiquier[nPos[i][1]][nPos[i][0]] = i + 1;

	for(j = 2; j < ( nWidth + 2) * 4 + 2; j++)
	{
		cout << "-";
		out << "-";
	}

	cout << "\n";
	out << "\n";

	for(i = 2; i < nHeight + 2; i++)
	{
		// sparateur initial
		cout << "| ";
		out << "| ";

		for(j = 2; j < nWidth + 2; j++)
		{
			cout << ( nEchiquier[i][j] < 10 ? "0" : "" ) << nEchiquier[i][j] << " | ";
			out << ( nEchiquier[i][j] < 10 ? "0" : "" ) << nEchiquier[i][j] << " | ";
		}

		cout << "\n";
		out << "\n";

		// --------
		for(j = 2; j < ( nWidth + 2) * 4 + 2; j++)
		{
			cout << "-";
			out << "-";
		}

		cout << "\n";
		out << "\n";
	}

	out.close();

	// libration des cases de l'echiquier
	for(i = 0; i < nHeight; i++)
		delete [] nEchiquier[i];
	delete [] nEchiquier;

	// libration du tableau des positions
	for(i = 0; i < nWidth * nHeight; i++)
		delete [] nPos[i];
	delete [] nPos;
}

/*
		void write_echiquier(int nWidth, int nHeight, int nCurrentX, int nCurrentY)
		Affichage de l'echiquier
		Paramtres:
			- nWidth: Largeur de l'echiquier
			- nHeight: Hauteur de l'echiquier
			- nStartX: Position X de dpart
			- nStartY: Position Y de dpart
		Retour: Aucun
*/
void write_echiquier(int nWidth, int nHeight, int nCurrentX, int nCurrentY)
{
	cout << "\n\n";
	// ---------------
	for(int j = 2; j < ( nWidth + 2) * 3 + 2; j++)
		cout << "-";

	cout << "\n";

	for(int i = 2; i < nHeight + 2; i++)
	{
		// sparateur initial
		cout << "| ";

		for(int j = 2; j < nWidth + 2; j++)
		{
			if(j == nCurrentX && i == nCurrentY)
				cout << "X | ";
			else
				cout << nEchiquier[i][j] << " | ";
		}

		cout << "\n";

		// --------
		for(int j = 2; j < ( nWidth + 2) * 3 + 2; j++)
			cout << "-";

		cout << "\n";
	}

	cout << "\n";
}